(function($) {

    RAVEN.namespace("RAVEN.TestExecution");

    /**
     * Move positions constants
     */
    var AFTER = "After";
    var BEFORE = "Before";

    /**
     * This Model defines the column layout for the given user
     */
    RAVEN.TestExecution.TestRunsTableMetaModel = Backbone.Model.extend({
        urlRoot: function() {
            return contextPath + '/rest/raven/1.0/testexec/' + this._getTestExecKey() + '/columns';
        },

        defaults: function() {
            return [];
        },

        _getTestExecKey: function () {
            return $('#raven-testexec-tests-issue-key').val();
        }
    });

    /**
     * This is the main view for the Test Runs table on a Test Execution issue
     */
    RAVEN.TestExecution.TestsView = Backbone.View.extend({
        el: '#raven-exec-tests',

        /**
         * Flag that indicates whether this view has finished rendering
         */
        ready: false,

        /**
         * DataTable Instance
         */
        dataTable: null,

        /**
         * The Test Execution issue key
         */
        testExecKey: null,

        /**
         * Check if issue is a Test Execution
         */
        isTestExecution: null,

        /**
         * Permissions
         */
        canExecute: false,
        canExecuteInline: false,
        canEdit: false,
        canAssign: false,

        /**
         * Test Run Statuses
         */
        testRunStatuses: null,

        /**
         * Current dropdown trigger element
         */
        currentDropdownTrigger: null,

        /**
         * The Bulk view
         */
        bulkEditView: null,

        /**
         * The filters view object
         */
        filtersView: null,

        /**
         * Current filters
         */
        filters: null,

        /**
         * The Test Runs toolbar view object
         */
        toolbarView: null,

        events: {

        },

        /**
         * Unscoped Events - these events must be registered elsewhere as they will not be
         * fired by backbone because they are not within the view element. Refer to "_delegateUnscopedEvents".
         */
        unscopedEvents: {
            'click.raven.testexec #raven-add-tests-link': '_onAddTestsClick',
            'click.raven.testexec #raven-add-testsets-link': '_onAddTestSetsClick',
            'click.raven.testexec a.raven-remove-command': '_onTestDeleteClick',
            'click.raven.testexec a.raven-bulk-remove-command': '_onBulkTestDeleteClick',
            'click.raven.testexec a.raven-rank-change-command': '_onChangeTestRankClick',
            'click.raven.testexec a.raven-assign-command': '_onAssignTestRunClick',
            'click.raven.testexec a.raven-bulk-assign-command': '_onBulkAssignTestRunClick',
            'click.raven.testexec a.raven-bulk-testexec-execute-inline': '_onBulkExecuteInlineTestClick',
            'click.raven.testexec .raven-testexec-execute-inline': '_onExecuteInlineTestClick',
            'aui-dropdown2-hide.raven.testexec .aui-dropdown2': '_onDownDownHide',
            'click.raven.testexec .raven-dropdown-trigger': '_onDropDownTriggerClick',
            'click.raven.testexec #fixTests' : '_fixTestsTriggerClick'
        },

        initialize: function() {
            var self = this;

            this._verifyTestsInTestPlan();

            this.model = new RAVEN.TestExecution.TestRunsTableMetaModel();

            this._delegateUnscopedEvents();

            AJS.messages.setup();

            this._clearMessage();

            this.testExecKey = $('#raven-testexec-tests-issue-key').val();
            this.isTestExecution = $('#raven-is-test-exec').val() === 'true';

            this.canExecute = $('#can-execute').val() === 'true';
            this.canExecuteInline = $('#can-execute-inline').val() === 'true';
            this.canEdit = $('#can-edit').val() === 'true';
            this.canAssign = $('#can-assign').val() === 'true';
            this.testRunStatuses = JSON.parse($('#test-run-statuses').val());

            this._setEditable(this.canEdit);

            $('#raven-exec-tests').block();
            this.model.fetch({
                success: function() {
                    $('#raven-exec-tests').unblock();
                    if (self.$el.closest(document.documentElement).length > 0) {
                        self.render();
                    }
                },
                error: function(model, response) {
                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.column-layout.error'), null, 'error');
                    $('#raven-exec-tests').unblock();
                }
            });
        },

        render: function() {
            this._clearMessage();
            
            // Render table skeleton
            var $table = RAVEN.TestExecution.renderTestRunsTable({'columns': _.values(this.model.attributes)});
            var $tableWrapper = this.$el.find('#testexec-tests-table-wrapper');
            $tableWrapper.empty();
            $tableWrapper.html($table);

            // render test run filters
            if (!this.filtersView) {
                this.filtersView = new RAVEN.TestExecution.TestRunFiltersView({
                    mainView: this
                });
            }

            // create the dataTable
            this._createTestRunsDataTable();


            // render bulk mode button
            if (this.canEdit === true && !this.bulkEditView) {
                this._renderBulkModeButton();
            }

            // render test runs toolbar view
            this._renderTestRunsToolbarView();

            $tableWrapper.show();

            this.$el.show();

            this.ready = true;

            return this;
        },

        destroy: function() {
            this._cleanView();

            // COMPLETELY UNBIND THE VIEW
            this.undelegateEvents();

            this._undelegateUnscopedEvents();

            this.$el.removeData().unbind();

            this.unbind();

            this.remove();
        },

        _cleanView: function() {
            if (this.dataTable) {
                this.dataTable.fnDestroy(true);
                this.dataTable = null;
            }

            this.ready = false;
            this.model = null;
            this.bulkEditView = null;

            var $tableWrapper = this.$el.find('#testexec-tests-table-wrapper');
            $tableWrapper.empty();
        },

        /**
         * Filters the datatable with the provided parameters
         */
        filter: function(filters) {
            this.filters = filters;

            this.dataTable.fnDraw();
        },

        /**
         * Resets all filters on datatable and redraws.
         */
        cleanFilters: function() {
            this.filters = null;

            var oSettings = this.dataTable.fnSettings();
            for(iCol = 0; iCol < oSettings.aoPreSearchCols.length; iCol++) {
                oSettings.aoPreSearchCols[ iCol ].sSearch = '';
            }
            oSettings.oPreviousSearch.sSearch = '';
            this.dataTable.fnDraw();
        },

        /**
         * Redraws the datatable
         */
        refresh: function() {
            this._cleanView();
            this.initialize();
        },

        /**
         * Handles the creation of the datatable.
         */
        _createTestRunsDataTable: function() {
            var self = this;

            var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + "/test";

            if(this.dataTable) {
                this.dataTable.fnDestroy();
                //$('#exec-entries-table').empty();
            }
            this.dataTable = $('#exec-entries-table').xrayDataTable({
                "oLanguage": {
                    "oPaginate": {
                        "sFirst": AJS.I18n.getText('common.label.first'),
                        "sLast": AJS.I18n.getText('common.label.last'),
                        "sNext": AJS.I18n.getText('common.label.next'),
                        "sPrevious": AJS.I18n.getText('common.label.previous')
                    },
                    "sInfoEmpty": AJS.I18n.getText('view-issue-precondition-tests-section.infoempty'),
                    "sEmptyTable": AJS.I18n.getText('view-issue-precondition-tests-section.emptytable'),
                    "sInfo": AJS.I18n.getText('common.table.label.showing-records'),
                    "sZeroRecords": AJS.I18n.getText('common.table.label.no-records-found'),
                    "sInfoFiltered": AJS.I18n.getText('common.table.label.table-filtered-from'),
                    "sSearch":  AJS.I18n.getText('common.table.label.search'),
                    "sLengthMenu": AJS.I18n.getText('common.table.label.show-menu-entries')
                },
                "sDom" : "l<t>ip",
                "bServerSide" : true,
                "sAjaxSource": path,
                "sAjaxDataProp": "entries",
                "bProcessing": false,
                "bAutoWidth": false,
                "sPaginationType": "full_numbers",
                "bSort": false,
                "aaSorting": [],
                "aoColumnDefs": this._createDataTableColumnDefs(),
                "fnServerData": function( sSource, aoData, fnCallback, oSettings){
                    self._fnServerData( sSource, aoData, fnCallback, oSettings );
                    if (self.bulkEditView) {
                        self.bulkEditView.setCheckBoxRowsSelectionValue(false);
                    }
                },
                "fnRowCallback": function( nRow, aData, iDisplayIndex, iDisplayIndexFull ) {
                    nRow.id = aData["key"];
                    $(nRow).data("testrun", aData["testrun"]);

                    $(nRow).addClass("testentry");

                    return nRow;

                },
                "fnDrawCallback": function( oSettings ) {
                    // set initial overflow to allow overflow columns to be visible
                    self.$el.css("overflow", "initial");

                    // hack for drag&drop to work
                    $('#exec-entries-table').closest('div').addClass('datatable-drag-overflow');

                    self._toogleTestsTable();
                    if (self.bulkEditView) {
                        self.bulkEditView.showHideBulkActions();
                    }

                    // remove loose tipsies
                    $('.tipsy:has(.testsets-custom-field-tooltip-close)').remove();

                    //add tipsy for non-executable tests
                    $(".not-executable").tipsy();
                },

                "fnInitComplete": function(settings, json) {
                	 $(".testexec-execution-button").parent().css("width","65px");
                	 $(".testexec-execution-button").parent().css("white-space","nowrap");

                }
            });

            if (this.canEdit === true) {
                this._initSortableTable("#exec-entries-table", false);
            }

            this._initSortableColumns();
        },

        /**
         * Creates the DataTable column def descriptors dynamically based on the user column layout
         */
        _createDataTableColumnDefs: function() {
            var self = this;
            var columns = _.values(this.model.attributes);

            var bulkToogleColumnTitle = null;
            var bulkActionColumnTitle = null;
            if (self.canEdit === true) {
                bulkToogleColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkSelectAllInput();

                var bulkActions = RAVEN.TestExecution.testRunBulkMenuActions({ 'canEdit' : self.canEdit, 'canAssign': self.canAssign, 'canExecuteInline': self.canExecuteInline, 'transitions': self.testRunStatuses});
                bulkActionColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkActionTrigger({id: "tests-bulk-action",actions: bulkActions});
            }

            var firstColumnDefs = [
                {
                    "mData": null,
                    "sTitle": bulkToogleColumnTitle,
                    "mDataProp": null,
                    "sWidth": "20px",
                    "sDefaultContent": "<input type='checkbox' ></input>",
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                        if (self.canEdit === true) {
                            return RAVEN.DataTable.BulkEdit.renderBulkSelectInputAndDragHandle({id: row["key"]});
                        } else {
                            return '';
                        }
                    },
                    "aTargets": [ 0 ]
                },
                {
                    "mData": "rank",
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                        return "<span class='rank' style='padding-left: 2px'>" + data + "</span>";
                    },
                    "aTargets": [ 1 ]
                }
            ];

            var userColumnDefs = this._createDataTableUserColumn(columns);

            var lastColumnDefs = [
                {
                    "mData": "status",
                    "sWidth" : "110px",
                    "mRender": function ( data, type, row ) {
                        return "<span data-status-id="+data.id+" class=\"aui-lozenge testentry-progress-bar\" style=\"background-color:"+data.color+"\" title=\""+data.description+"\">"+data.name+"</span>";
                    },
                    "aTargets": [ 2 + columns.length ]
                },
                {
                    "mData": null,
                    "sTitle": bulkActionColumnTitle,
                    "sWidth": "114px",
                    "sClass" : "testexec-testruns-lastcol-actions",
                    "mRender": function ( data, type, row ) {
                        var hasStartedExecuting = row["startedExecuting"];
                        var canExecuteTestRun = row["canExecute"];
                        var htmlMenus = '';

                    	if(self.canAssign === true) {
                            var actionsMenu = RAVEN.TestExecution.actionsMenu({
                                key: row["key"],
                                canEdit: self.canEdit,
                                canAssign: self.canAssign,
                                rank: row["rank"]
                            });

                            htmlMenus += actionsMenu;
                        }

                        var testRunMenu = RAVEN.TestExecution.testRunMenu({
                            key: row["key"],
                            canExecute: canExecuteTestRun && self.canExecute,
                            hasStartedExecuting: hasStartedExecuting,
                            canExecuteInline: canExecuteTestRun && self.canExecute && self.canExecuteInline,
                            executeHref: self._getExecuteTestHref(row["key"]),
                            transitions: row["transitions"],
                            isFinal: row["status"].isFinal
                        });

                        htmlMenus += testRunMenu;

                        if(canExecuteTestRun === false) {
                            var str = row["disallowedExecuteMessage"];
                            var element = '<a class="not-executable" title="'+str +'"></a>';
                            htmlMenus += element;
                        }

                    	return htmlMenus;

                    },
                    "bSortable": false,
                    "bSearchable": false,
                    "aTargets": [ 3 + columns.length]
                }
            ];

            return firstColumnDefs.concat(userColumnDefs).concat(lastColumnDefs);
        },

        /**
         * Creates the DataTable column def for user columns
         */
        _createDataTableUserColumn: function(columns) {
            return _.map(columns, function(column, index) {
                return {
                    "mData": 'userColumns.' + column.id,
                    "bSortable": false,
                    "bSearchable": false,
                    "aTargets": [ 2 + index ]
                };
            });
        },

        message: function (title, message, type) {
            type = (type === undefined || type == "") ? "error" : type;

            if (!message && title) {
                message = title;
                title = null;
            }

            if (type === "error") {
                this._clearMessage();
                AJS.messages.error('#raven-exec-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "warning") {
            	this._clearMessage();
    			AJS.messages.warning('#raven-exec-tests-message-bar', {
    				title : title,
    				body : message
    			});
    		}

            if (type === "success") {
                AJS.messages.success('#raven-exec-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "info") {
                AJS.messages.info('#raven-exec-tests-message-bar', {
                    title: title,
                    body: message
                });
            }
        },

        _clearMessage: function() {
            $('#raven-exec-tests-message-bar').empty();
        },

        _fnServerData: function( sSource, aoData, fnCallback, oSettings ) {
            var self = this;

            if (this.filters) {
                aoData = aoData.concat(this.filters);
            }

            $('#raven-exec-tests').block();
            oSettings.jqXHR = $.ajax( {
                "dataType": 'json',
                "type": "GET",
                "url": sSource,
                "data": aoData,
                "error": function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }
                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.processingerror'), errorThrown, 'error');
                },
                "complete": function() {
                    $('#raven-exec-tests').unblock();
                },
                "success": fnCallback
            });
        },

        _updateHistory: function() {
            if ($('#changehistory-tabpanel').attr('class') == 'active') {
                var alt_token = $('input[name="atl_token"]').val();
                $.ajax({
                    type : "GET",
                    url : contextPath + '/browse/' + this.testExecKey +"?page=com.atlassian.jira.plugin.system.issuetabpanels:changehistory-tabpanel&_pjax=true&_="+alt_token,
                    contentType : "text/plain",
                    success : function (data, status) {
                        var tab = $(data).find('#activitymodule div.mod-content #issue_actions_container');
                        var jiraTab = $('#activitymodule div.mod-content').find('#issue_actions_container');
                        $(jiraTab).html(tab);
                    }
                });
            }
        },

        _updateDates: function () {
            var updatedDate = null;

            if(AJS.$('#datesmodule #updated-date').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-date time.livestamp');
            } else if(AJS.$('#datesmodule #updated-val').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-val time.livestamp');
            } else {
                return null;
            }

            $.ajax({
                type : "GET",
                url : AJS.contextPath() + '/rest/raven/1.0/date/' + JIRA.Issue.getIssueKey() + "/updated",
                contentType : "application/json",
                success : function (data) {
                    var title = data.title;
                    var dateTimeIso8601 = data.dateTimeIso8601;
                    updatedDate.attr("datetime", dateTimeIso8601);
                    $.each(AJS.$('#datesmodule dd.date'), function() {
                        if($(this).find('#updated-date time.livestamp').length > 0 || $(this).find('#updated-val time.livestamp').length > 0) {
                            $(this).attr("title", title);
                        }
                    });
                    updatedDate.livestamp();
                }
            });
        },

        _setEditable: function(editable) {
            if (editable === true) {
                $('#add-tests').show();
            } else {
                $('#add-tests').hide();
            }
        },

        _setExecutable: function(executable) {
            if (executable === true) {
                $('.execute-command').show();
            } else {
                $('.execute-command').hide();
            }
        },

        _toogleTestsTable: function() {
            if (this.dataTable && this.dataTable.fnRecordsTotal() > 0) {
                $('#raven-exec-tests .notests').hide();
                $('#exec-tests-section').show();
            } else {
                $('#exec-tests-section').hide();
                $('#raven-exec-tests .notests').show();
            }
        },

        _addTests: function(testKeys, assignee, isToIgnoreDisallowedTestStatuses) {
            var self = this;

            $('#raven-exec-tests').block();

            var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + '/test';

            if (!assignee || assignee == "-1") {
                assignee = null;
            }

            var value = {
                "keys": testKeys,
                "assignee": assignee,
                "ignoreDisallowedTestStatuses" : isToIgnoreDisallowedTestStatuses
            };
            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                timeout: 60000, // sets timeout to 60 seconds
                data: JSON.stringify(value),
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.add.error', self.testExecKey), errorThrown);
                    $('#raven-exec-tests').unblock();
                },
                success: function(data, textStatus, jqXHR) {
                    if (data.errors && data.errors.length > 0) {
                        var errorMsg = '';
                        $.each($(data.errors), function(i, msg) {
                            errorMsg = errorMsg + msg;
                            if (i < data.errors.length) {
                                errorMsg = errorMsg + '<br/>';
                            }
                        });
                        self.message(AJS.I18n.getText('view-issue-testexec-tests-section.add.controlled.error', self.testExecKey), errorMsg);
                    }

                    $('#raven-exec-tests').unblock();
                    self._processAddResults(data);
                    self._verifyTestsInTestPlan();
                },
                complete: function() {
                    self._updateHistory();
                    self._updateDates();
                }
            });
        },

        _processAddResults: function(data) {
            // new tests
            if (data.tests && data.tests > 0) {
                this.dataTable.fnDraw();
                //this.dataTable.fnAddData(data.tests);

                // Thrigger disabled due to unwanted flickering effect
                // JIRA.trigger(JIRA.Events.REFRESH_ISSUE_PAGE, [JIRA.Issue.getIssueId()]);
            }
            // update status progress bar
            if (data.statusbar) {
                this._updateExecutionStatus(data.statusbar);
            }
        },

        _deleteTest: function(testKeys) {
            var self = this;

            $('#raven-exec-tests').block();

            //if we have records to delete
            if(testKeys && testKeys.length > 0) {
                //Build query param
                var testKeysParam = this._buildQueryParamsFromSelectedRows(testKeys);

                $.ajax({
                    type: "DELETE",
                    url: contextPath + '/rest/raven/1.0/testexec/' + self.testExecKey + "/test/?" + testKeysParam + "" ,
                    processData: true,
                    async: true,
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }
                        if (testKeys.length == 1) {
                            self.message(AJS.I18n.getText('view-issue-testexec-tests-section.delete.error', testKeys[0]), errorThrown);
                        } else {
                            self.message(AJS.I18n.getText('view-issue-testexec-tests-section.bulk.delete.error'), errorThrown);
                        }
                        $('#raven-exec-tests').unblock();
                    },
                    success: function(data, textStatus, jqXHR) {
                        $('#raven-exec-tests').unblock();
                        self._deleteTestsCallback(testKeys, data);
                    },
                    complete: function() {
                        self._updateHistory();
                        self._updateDates();
                    }
                });
            }
        },

        _deleteTestsCallback: function(testKeys, data) {
            var self = this;
            if (testKeys && testKeys.length > 0) {
                var allData = this.dataTable.fnGetData();
                if (allData.length >= 1) {
                    _.each(_.initial(testKeys), function(testKey) {
                        self.dataTable.fnDeleteRow(testKey, null, false);
                    });
                    this.dataTable.fnDeleteRow(_.last(testKeys));
                } else {
                    // don't refresh the table if the last element is removed because it will request more data from server
                    this.dataTable.fnDeleteRow(_.last(testKeys), null, false);
                }
            }

            // update status progress bar
            if (data.statusbar) {
                this._updateExecutionStatus(data.statusbar);
            }

            // Thrigger disabled due to unwanted flickering effect
            // JIRA.trigger(JIRA.Events.REFRESH_ISSUE_PAGE, [JIRA.Issue.getIssueId()]);
        },

        _moveTest: function(table, movedTestData, ancorTestData, position) {
            var self = this;

            $('#raven-exec-tests').block();

            var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + '/test/' + movedTestData.key + '/move'

            var value = {};
            if (position == AFTER) {
                value.after = '/' + ancorTestData.key;
            } else if (position == BEFORE) {
                value.before = '/' + ancorTestData.key;
            } else {
                value.index = position;
            }
            var data = JSON.stringify(value);

            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                data: data,
                async: true,
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }
                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.move.error', movedTestData.key), errorThrown);

                    $('#raven-exec-tests').unblock();
                },
                success: function(data, textStatus, jqXHR) {
                    $('#raven-exec-tests').unblock();
                    table.fnDraw(false);
                }
            });
        },

        _initSortableColumns: function() {
            var self = this;

            $('#exec-entries-table thead tr').sortable({
                items: 'th:not(.fixed)',
                placeholder: "dragging-placeholder",
                update: function (event, ui) {
                    var ancorNode = ui.item.prev();
                    var columnId = $(ui.item).data('column-id');
                    if (columnId) {
                        var anchorColumnId = null;
                        if (!$(ancorNode).hasClass('fixed')) {
                            anchorColumnId = $(ancorNode).data('column-id');
                        }

                        self.columnLayoutView.moveColumn(columnId, anchorColumnId);
                    }
                }
            });
        },

        /**
         * Adds support for moving datatable items
         */
        _initSortableTable: function(tableSelector) {
            var self = this;

            $(tableSelector).sortable({
                items:'tbody tr',
                handle:'.draghandle',
                placeholder: "drag-placeholder",
                helper: function(e, tr) {
                    var $originals = tr.children();
                    var $helper = tr.clone();
                    $helper.children().each(function(index) {
                        // Set helper cell sizes to match the original sizes
                        $(this).width($originals.eq(index).width())
                    });
                    return $helper;
                },
                update: function (event, ui) {
                    var position = null;
                    var ancorNode = null;

                    var ancorNode = ui.item.prev();
                    if (ancorNode.length > 0) {
                        position = AFTER;
                    } else {
                        var ancorNode = ui.item.next();
                        if (ancorNode.length > 0) {
                            position = BEFORE;
                        }
                    }

                    if (position != null) {
                        var movedTestData = self.dataTable.fnGetData(ui.item[0]);
                        var ancorTestData = self.dataTable.fnGetData(ancorNode[0]);

                        if (movedTestData && movedTestData.key && ancorTestData && ancorTestData.key) {
                            self._moveTest(self.dataTable, movedTestData, ancorTestData, position);
                        }
                    }
                }
            });
        },

        /**
         * Updates the overall execution status bar.
         * TODO: this should also be a Backbone component
         */
        _updateExecutionStatus: function(statusbar) {
            var statusContent = RAVEN.TestExecution.execProgressBar({statusbar: statusbar});
            var pb = $('#exec-tests-progressbar');
            pb.empty();
            pb.append(statusContent);
        },

        /**
         * Bulk Inline execution of Test Runs.
         */
        _bulkInlineExecuteTestRun: function(rows, statusId) {
            var self = this;

            $('#raven-exec-tests').block();

            var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + '/execute';

            var testKeys = _.map(rows, function(row) {
                return row.id;
            });

            var value = {
                keys: testKeys,
                status: statusId.toString()
            };

            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                data: JSON.stringify(value),
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.inline-execute.error', self.testExecKey), errorThrown);
                },
                success: function(data, textStatus, jqXHR) {
                    _.each(data.obj.tests, function(test) {
                        var idObject = '#' + test.key;
                        var checkBox = $(idObject + ' :checkbox');
                        var valueCheckBox = checkBox.is(':checked');

                        var $row = $(idObject);

                        var index = undefined;
                        var statusColumn = self.dataTable.find('thead th[data-column-id="testrun_status"]');
                        if (statusColumn.length > 0) {
                            index = $.inArray(statusColumn[0], self.dataTable.find('thead th'));
                        }

                        self.dataTable.fnUpdate(test, $row[0], index, false, false);

                        // refresh the bulk option and force unselection of row due to redrawning of row
                        if (self.bulkEditView) {
                            self.bulkEditView.showHideBulkActions();
                            self.bulkEditView.setCheckBoxRowsSelectionValue(valueCheckBox, idObject.substr(1, idObject.length));
                        }
                    });

                    self._updateExecutionStatus(data.obj.statusbar);

                    if(data.errors.length > 0){
                        var finalErrors = "";
                        for(var i = 0 ; i < data.errors.length ; i++){
                            finalErrors += data.errors[i] + "\n";
                        }
                        self.message(AJS.I18n.getText('test-execution.workflow-status.error-message.title'), finalErrors, "warning");
                    }

                    if(data.statusChanged){
                        JIRA.trigger(JIRA.Events.REFRESH_ISSUE_PAGE, [JIRA.Issue.getIssueId()]);
                    }
                },
                complete: function() {
                    $('#raven-exec-tests').unblock();
                }
            });
        },

        /**
         * Inline execution of Test Runs.
         */
        _inlineExecuteTestRun: function($row, statusId) {
            var self = this;

            $('#raven-exec-tests').block();

            var rowData = this.dataTable.fnGetData($row[0]);

            if (rowData.status.id !== statusId) {
                var testKey = rowData.key;

                var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + '/execute/' + testKey;

                $.ajax({
                    type: "POST",
                    url: path,
                    processData: true,
                    async: true,
                    data: statusId.toString(),
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }

                        self.message(AJS.I18n.getText('view-issue-testexec-tests-section.inline-execute.error', self.testExecKey), errorThrown);
                    },
                    success: function(data, textStatus, jqXHR) {

                        var idObject = '#'+$row[0].id;
                        var checkBox = $(idObject + ' :checkbox');
                        var valueCheckBox = checkBox.is(':checked');

                        var index = undefined;
                        var statusColumn = self.dataTable.find('thead th[data-column-id="testrun_status"]');
                        if (statusColumn.length > 0) {
                            index = $.inArray(statusColumn[0], self.dataTable.find('thead th'));
                        }

                        self.dataTable.fnUpdate(data.obj.test, $row[0], index, false, false);

                        self._updateExecutionStatus(data.obj.statusbar);

                        // refresh the bulk option and force unselection of row due to redrawning of row
                        if(self.bulkEditView){
	                        self.bulkEditView.showHideBulkActions();
	                        self.bulkEditView.setCheckBoxRowsSelectionValue(valueCheckBox, idObject.substr(1, idObject.length));
	                    }

                        if(data.errors.length > 0){
                        	var finalErrors = "";
                        	for(var i = 0 ; i < data.errors.length ; i++){
                        		finalErrors += data.errors[i] + "\n";
                        	}
                        	self.message(AJS.I18n.getText('test-execution.workflow-status.error-message.title'), finalErrors, "warning");
                        }

                        if(data.statusChanged){
                        	JIRA.trigger(JIRA.Events.REFRESH_ISSUE_PAGE, [JIRA.Issue.getIssueId()]);
                        }
                    },
                    complete: function() {
                        $('#raven-exec-tests').unblock();
                    }
                });
            }
        },

        _updateRankRow: function(newRankValue, originalRankSpanValue, row) {
            //Validate values
            if(newRankValue > 0 && newRankValue <= this.dataTable.fnSettings().fnRecordsTotal() && newRankValue != originalRankSpanValue) {
                var movedTestData = this.dataTable.fnGetData(row);

                if (movedTestData && movedTestData.key) {
                    this._moveTest(this.dataTable, movedTestData, null, newRankValue);
                }
            }
        },

        /**
         * Build query param
         */
        _buildQueryParamsFromSelectedRows: function(selectedIds) {
            var testKeys = "";
            var i;
            for (i = 0; i < selectedIds.length; ++i) {
                if(testKeys.length > 0){
                        testKeys+="&";
                }
                testKeys += "testKeys=" + selectedIds[i];
            }

            return testKeys;
        },

        _getExecuteTestHref: function(testKey) {
            var link = contextPath + '/secure/XrayExecuteTest!default.jspa?testExecIssueKey=' + this.testExecKey;
            link += '&testIssueKey=' + testKey;
            return link;
        },

        _confirmationChangeRank: function(title, message, row, originalRankSpanValue, yesFunction) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var innerHtml = RAVEN.Dialog.editValueDialog({title:title, fieldName:message, actionText:AJS.I18n.getText('common.action.ok')});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var newRankValue = $('#edit-value-dialog-value')[0].value;
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(newRankValue, originalRankSpanValue, row);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        _confirmationDeleteTest: function(el, yesFunction, bulk) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var title = null;
                    var message = null;
                    if(self.isTestExecution === true) {
                        if (bulk === true) {
                            title = AJS.I18n.getText('view-issue-testexec-tests-section.bulk.delete.confirm.title');
                            message = AJS.I18n.getText('view-issue-testexec-tests-section.bulk.delete.confirm.message', self.testExecKey);
                        } else {
                            title = AJS.I18n.getText('view-issue-testexec-tests-section.delete.confirm.title', el);
                            message = AJS.I18n.getText('view-issue-testexec-tests-section.delete.confirm.message', self.testExecKey);
                        }
                    } else {
                        if (bulk === true) {
                            title = AJS.I18n.getText('view-issue-testexec-tests-section.bulk.delete.confirm.title');
                            message = AJS.I18n.getText('view-issue-subtask.testexec-tests-section.bulk.delete.confirm.message', self.testExecKey);
                        } else {
                            title = AJS.I18n.getText('view-issue-testexec-tests-section.delete.confirm.title', el);
                            message = AJS.I18n.getText('view-issue-subtask.testexec-tests-section.delete.confirm.message', self.testExecKey);
                        }
                    }
                    var severity = "info";
                    var innerHtml = RAVEN.Dialog.deleteConfirmationDialog({title:title, message:message, severity:severity});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(el);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        _renderBulkModeButton: function() {
            var $wrapper = this.$el.find('.xrayDataTables_wrapper');
            // if filter already exists then return
            if ($wrapper.find('#bulk-edit-action').length > 0) {
                return;
            }

            this.bulkEditView = new RAVEN.DataTable.BulkEditView({
                el: $wrapper
            });
        },

        _renderTestRunsToolbarView: function() {
            var $wrapper = this.$el.find('.xrayDataTables_wrapper');
            // if filter already exists then return
            if ($wrapper.find('#toolbar-wrapper').length > 0) {
                return;
            }

            var toolbarHtml = RAVEN.TestExecution.renderTestRunsToolBar();
            $wrapper.find('.xrayDataTables_length').after(toolbarHtml);

            this.columnLayoutView = new RAVEN.TestExecution.TestRunsColumnLayoutView({
                mainView: this
            });
        },

        _assignTestRun: function(username, rows) {
            var self = this;

            $('#raven-exec-tests').block();

            var path = contextPath + '/rest/raven/1.0/testexec/' + this.testExecKey + '/assign';

            var testKeys = _.map(rows, function(row) {
                return row.id;
            });

            var value = {
                keys: testKeys,
                assignee: !username || username == -1 ? null : username
            };
            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                data: JSON.stringify(value),
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.assign.error'), errorThrown);
                },
                success: function(data, textStatus, jqXHR) {
                    self._onAssignOperationSuccess(data, rows);
                },
                complete: function() {
                    $('#raven-exec-tests').unblock();
                    self._updateHistory();
                }
            });
        },

        _onAssignOperationSuccess: function(data, rows) {
            var self = this;

            if (data && data.length > 0 && data.length === rows.length) {
                var index = -1;
                var assigneeColumn = self.dataTable.find('thead th[data-column-id="testrun_assignee"]');
                if (assigneeColumn.length > 0) {
                    index = $.inArray(assigneeColumn[0], self.dataTable.find('thead th'));
                }

                _.each(data, function(test, i) {
                    if (index > -1) {
                        if (test.assignee) {
                            var userLink = RAVEN.TestExecution.renderUserLink({ user: test.assignee, context: contextPath});
                            self.dataTable.fnUpdate(userLink, rows[i], index, false, false);
                        } else {
                            self.dataTable.fnUpdate("", rows[i], index, false, false);
                        }
                    }

                    // set new row data assignee object
                    var rowData = self.dataTable.fnGetData(rows[i]);
                    if (rowData) {
                        rowData.assignee = test.assignee;
                    }
                });

                // TODO: assignee filter
            }
        },

        /****************************************************************
         *                        EVENT HANDLERS
         ****************************************************************/

        _delegateUnscopedEvents: function() {
            var self = this;
            $(document).off('.raven.testexec');

            $.each(this.unscopedEvents, function(key, val) {
                var eventName = key.substring(0, key.indexOf(' '));
                var eventTrigger = key.substring(key.indexOf(' ') + 1);

                $(document).on(eventName, eventTrigger, $.proxy(self[val], self));
            });
        },

        _undelegateUnscopedEvents: function() {
            $(document).off('.raven.testexec');
        },

        /**
         * Add tests button click handler
         */
        _onAddTestsClick: function() {
            var self = this;

            this._clearMessage();

            var labels = {};
            labels.label = AJS.I18n.getText('dialogs.issueselector.test.label');
            labels.title = AJS.I18n.getText('dialogs.issueselector.test.title', this.testExecKey);
            labels.description = AJS.I18n.getText('dialogs.issueselector.test.description');
            labels.enter = AJS.I18n.getText('dialogs.issueselector.test.enter');

            var issueTypeId = $('#test-issue-type-id').val();

            new RAVEN.Dialogs.createIssueSelectorAdvancedDialog('TEST', 'key not in testExecutionTests(' + this.testExecKey + ')', issueTypeId, this.testExecKey, labels, function(issues, assignee) {
                if (issues && issues.length > 0) {
                    self._addTests(issues, assignee);
                }
            }, null, self.canAssign, true);
        },

        /**
         * Add test sets button click handler
         */
        _onAddTestSetsClick: function() {
            var self = this;

            this._clearMessage();

            var labels = {};
            labels.label = AJS.I18n.getText('dialogs.issueselector.testset.label');
            labels.title = AJS.I18n.getText('dialogs.issueselector.testset.title', this.testExecKey);
            labels.description = AJS.I18n.getText('dialogs.issueselector.testset.description');
            labels.enter = AJS.I18n.getText('dialogs.issueselector.testset.enter');

            var issueTypeId = $('#testset-issue-type-id').val();

            new RAVEN.Dialogs.createIssueSelectorAdvancedDialog('TESTSET', 'key not in testSetFullyIn(' + this.testExecKey + ')', issueTypeId, this.testExecKey, labels, function(issues, assignee) {
                if (issues && issues.length > 0) {
                    self._addTests(issues, assignee, $('#cbHideForbiddenTests').is(':checked'));
                }
            }, null, self.canAssign, true);
        },

        /**
         * Delete test click handler
         */
        _onTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            var testKey = $(e.target).closest('a').data('key');
            this._confirmationDeleteTest([ testKey ], this._deleteTest);
        },

        /**
         * Delete Bulk test click handler
         */
        _onBulkTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            if(this.bulkEditView.getSelectedRowIds().length > 0) {
                this._confirmationDeleteTest(this.bulkEditView.getSelectedRowIds(), this._deleteTest, true);
            }
        },

        /**
         * Change Test rank click handler.
         */
        _onChangeTestRankClick: function(e) {
            var originalRankSpanValue = $(e.target).attr('data-key');

            var $row = $(this.currentDropdownTrigger).closest('tr');

            e.stopPropagation();
            e.preventDefault();
            this._clearMessage();

            var rowKey = this.dataTable.fnGetData($row[0]).key;

            var label = AJS.I18n.getText('dialogs.edit-value.rank.label');
            var title = rowKey + ' - ' + AJS.I18n.getText('dialogs.edit-value.rank.title');
            this._confirmationChangeRank(title, label, $row[0], originalRankSpanValue, this._updateRankRow);
        },

        /**
         * Assign Test Run click handler
         */
        _onAssignTestRunClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            var self = this;

            var $row = $(this.currentDropdownTrigger).closest('tr');

            var testData = this.dataTable.fnGetData($row[0]);

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var innerHtml = RAVEN.TestExecution.renderUserPickerDialog({
                        currentAssignee: testData.assignee,
                        issueKey: self.testExecKey,
                        contextPath: contextPath
                    });
                    callback(innerHtml);

                    var control = new JIRA.AssigneePicker({
                        element: $('.xray-assignee-picker')
                    });
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var assignee = $('#raven-user-picker')[0].value;
                    var proxyFunc = $.proxy(self._assignTestRun, self);
                    if(assignee === "___unassigned___") {
                    	assignee = "\u241Eunassigned\u241E";
                    }
                    proxyFunc(assignee, [ $row[0] ]);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        /**
         * Bulk Assign Test Runs click handler
         */
        _onBulkAssignTestRunClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            var self = this;

            var rows = this.bulkEditView.getSelectedRows();

            if(rows.length > 0) {
                var dialog = new JIRA.FormDialog({
                    content: function (callback) {
                        var innerHtml = RAVEN.TestExecution.renderUserPickerDialog({
                            currentAssignee: null,
                            issueKey: self.testExecKey,
                            contextPath: contextPath
                        });
                        callback(innerHtml);

                        AJS.params.assigneeEditIssueKey = self.testExecKey;
                        var control = new JIRA.AssigneePicker({
                            element: $('.xray-assignee-picker')
                        });
                    },
                    submitHandler: function (e) {
                        e.preventDefault();
                        var assignee = $('#raven-user-picker')[0].value;
                        var proxyFunc = $.proxy(self._assignTestRun, self);
                        if(assignee === "___unassigned___") {
                        	assignee = "\u241Eunassigned\u241E";
                        }
                        proxyFunc(assignee, rows);
                        dialog.hide();
                    },
                    autoClose : true
                });
                dialog.show();
            }
        },

        /**
         * Bulk Execute inline test click handler
         */
        _onBulkExecuteInlineTestClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            var self = this;

            var rows = this.bulkEditView.getSelectedRows();
            var statusId = $(e.target).data('status-id');

            if(rows.length > 0) {
                self._bulkInlineExecuteTestRun(rows, statusId);
            }
        },

        /**
         * Execute inline test click handler
         */
        _onExecuteInlineTestClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            if (this.currentDropdownTrigger) {
                var $row = $(this.currentDropdownTrigger).closest('tr');
                var statusId = $(e.target).data('status-id');

                if ($row.length > 0) {
                    this._inlineExecuteTestRun($row, statusId);
                }
            }

            this.currentDropdownTrigger = null;

            return false;
        },

        /**
         * Save current dropdown trigger
         */
        _onDropDownTriggerClick: function(e) {
            this.currentDropdownTrigger = $(e.target);
        },

        /**
         * Clear current dropdown trigger
         */
        _onDownDownHide: function() {
        },

        _verifyTestsInTestPlan: function() {
            $("#raven-fix-test-testplan-message-bar").empty();
        	$.ajax({
                type : "GET",
                url : contextPath + '/rest/raven/1.0/testexec/' + $("#raven-testexec-tests-issue-key").val() + '/validateTestPlans',
                contentType : "application/json",
                success : function (data, status) {
                    data.forEach(function(data){
                    	AJS.messages.warning('#raven-fix-test-testplan-message-bar', {
            				title : AJS.I18n.getText('view-issue-testexec-tests-section.validate-test-plans.testNotInTestPlan.title'),
            				body : data
            			});
                    });
                }
            });
        },

        _fixTestsTriggerClick: function(e) {
        	var testplan = $(e.target).attr("data-testplan");
    		var tests = $(e.target).attr("data-tests");
    		$.ajax({
                type : "POST",
                url : contextPath + '/rest/raven/1.0/testexec/' + $("#raven-testexec-tests-issue-key").val() + '/addAllTestsToTestPlan?testPlan=' + testplan,
                data : tests,
                contentType : "application/json",
                success : function (data, status) {
                	// close message
                	$(e.target.parentElement).remove();
                }
            });
        }

    });

    RAVEN.TestExecution.TestsView.REFRESH_VIEW = "refreshTestsView";

})(AJS.$);
