(function($) {

    RAVEN.namespace("RAVEN.TestPlan");

    /**
     * Move positions constants
     */
    var AFTER = "After";
    var BEFORE = "Before";

    /**
     * This Model defines the column layout for the given user
     */
    RAVEN.TestPlan.TestsTableMetaModel = Backbone.Model.extend({
        urlRoot: function() {
            return contextPath + '/rest/raven/1.0/testplan/' + this._getTestPlanKey() + '/columns';
        },

        defaults: function() {
            return [];
        },

        _getTestPlanKey: function () {
            return $('#raven-testplan-tests-issue-key').val();
        }
    });

    /**
     * This is the main view for the Tests table on a Test Plan issue
     */
    RAVEN.TestPlan.TestsView = Backbone.View.extend({
        el: '#raven-plan-tests',

        /**
         * Flag that indicates whether this view has finished rendering
         */
        ready: false,

        draggable: true,

        /**
         * DataTable Instance
         */
        dataTable: null,

        /**
         * The Test Plan issue key
         */
        testPlanKey: null,

        /**
         * Permissions
         */
        canExecute: false,
        canExecuteInline: false,
        canEdit: false,
        canAssign: false,

        /**
         * Current dropdown trigger element
         */
        currentDropdownTrigger: null,

        /**
         * The Bulk view
         */
        bulkEditView: null,

        /**
         * The filters view object
         */
        filtersView: null,

        /**
         * Current filters
         */
        filters: null,

        /**
         * The Tests toolbar view object
         */
        toolbarView: null,

        /**
         * Whether using Test hierarchy or not
         */
        isTestRepositoryEnabled: false,

        /**
         * The Subtle Single Select Environment Picker
         */
        environmentPicker: null,

        /**
         * The Environments available
         */
        environments: null,

        /**
         * The Analysis Environment
         */
        analysisEnvironment: null,

        /**
         * The Test ids of the expanded inner tables
         */
        innerTableIds: null,

        events: {
            'update-tests' : '_updateTestsHandler'
        },

        /**
         * flag to indicate if test exec was created previously
         * */
        flagCreatedTestExec : false,

        /**
         * stores the reference to the warning flag when creating a test exec
         * */
        warningMessage : null,
        /**
         * Unscoped Events - these events must be registered elsewhere as they will not be
         * fired by backbone because they are not within the view element. Refer to "_delegateUnscopedEvents".
         */
        unscopedEvents: {	// TODO all these need to be revisited and check if needed
            'click.raven.testplan #raven-plan-add-tests-link': '_onAddTestsClick',
            'click.raven.testplan #raven-plan-add-testsets-link': '_onAddTestSetsClick',
            'click.raven.testplan a.raven-remove-command': '_onTestDeleteClick',
            'click.raven.testplan a.raven-remove-command-test-run': '_onTestRunDeleteClick',
            'click.raven.testplan a.raven-bulk-remove-command': '_onBulkTestDeleteClick',
            'click.raven.testplan a.raven-bulk-create-command': '_onBulkCreateTestExecutionClick',
            'click.raven.testplan a#raven-plan-create-te-link': '_onCreateTestExecutionClick',
            'click.raven.testplan a#raven-plan-create-te-status-link': '_onCreateTestExecutionWithStatusClick',
            'click.raven.testplan a.raven-assign-command': '_onAssignTestRunClick',
            'click.raven.testplan .raven-testplan-execute-inline': '_onExecuteInlineTestClick',
            'click.raven.testplan .raven-dropdown-trigger': '_onDropDownTriggerClick',
            'aui-dropdown2-hide.raven.testexec .aui-dropdown2': '_onDownDownHide',
            'click.raven.testplan #plan-entries-table td .raven_expand' : '_onExpandClick',
            'click.raven.testplan a.testplan-raven-rank-change-command': '_onChangeTestRankClick',
        },

        initialize: function() {
            var self = this;

            this.model = new RAVEN.TestPlan.TestsTableMetaModel();

            this.isTestRepositoryEnabled = $('#raven-test-repository-enabled').val() === "true";

            this._delegateUnscopedEvents();

            AJS.messages.setup();

            this._clearMessage();

            this.testPlanKey = $('#raven-testplan-tests-issue-key').val();

            this.canExecute = $('#can-execute').val() === 'true';
            this.canExecuteInline = $('#can-execute-inline').val() === 'true';
            this.canEdit = $('#can-edit').val() === 'true';
            this.canAssign = $('#can-assign').val() === 'true';
            this._setEditable(this.canEdit);
            this._setEnvironments();
            this._initTestRunEvent();


            $('#raven-plan-tests').block();
            this.model.fetch({
                success: function() {
                    $('#raven-plan-tests').unblock();
                    if (self.$el.closest(document.documentElement).length > 0) {
                        self.render();
                    }
                },
                error: function(model, response) {
                    self.message(AJS.I18n.getText('view-issue-testplan-tests-section.column-layout.error'), null, 'error');
                    $('#raven-plan-tests').unblock();
                }
            });
        },

        render: function() {
            var self = this;
            
            this._clearMessage();

            // Render table skeleton
            var $table = RAVEN.TestPlan.renderTestsTable({'columns': _.values(this.model.attributes), isTestRepositoryEnabled: this.isTestRepositoryEnabled});
            var $tableWrapper = this.$el.find('#testplan-tests-table-wrapper');
            $tableWrapper.empty();
            $tableWrapper.html($table);

            // render test run filters
            if (!this.filtersView) {
                this.filtersView = new RAVEN.TestPlan.TestFiltersView({
                    mainView: this
                });
            }

            // create the dataTable
            this._createTestsDataTable();


            // render bulk mode button
            if (this.canEdit === true && !this.bulkEditView) {
                this._renderBulkModeButton();
            }

            // render test runs toolbar view
            this._renderTestsToolbarView();

            this.environmentPicker = this._createSubtleSingleSelect($('#raven-analysis-environment-select'));
            $(self.environmentPicker).on('selectedItemChanged', function(e) {
                self.analysisEnvironment = self.environmentPicker.getSelectedItem().id;
                self.innerTableIds = [];
                $("tr.testplanInnerTable").each(function( ) {
                    self.innerTableIds.push($(this).prev().attr("id"));
                });
                self.dataTable.fnDraw()
            });

            $tableWrapper.show();

            this.$el.show();

            this.ready = true;

            return this;
        },

        destroy: function() {
            this._cleanView();

            // COMPLETELY UNBIND THE VIEW
            this.undelegateEvents();

            this._undelegateUnscopedEvents();

            this.$el.removeData().unbind();

            this.unbind();

            this.remove();
        },

        _cleanView: function() {
            if (this.dataTable) {
                this.dataTable.fnDestroy(true);
                this.dataTable = null;
            }

            this.ready = false;
            this.model = null;
            this.bulkEditView = null;

            var $tableWrapper = this.$el.find('#testplan-tests-table-wrapper');
            $tableWrapper.empty();
        },

        /**
         * Filters the datatable with the provided parameters
         */
        filter: function(filters) {
            var self = this;
            self.filters = filters;

            self.innerTableIds = [];
            $("tr.testplanInnerTable").each(function( ) {
                self.innerTableIds.push($(this).prev().attr("id"));
            });

            self.dataTable.fnDraw();
        },

        /**
         * Resets all filters on datatable and redraws.
         */
        cleanFilters: function() {
            this.filters = null;
            this.innerTableIds = null;

            var oSettings = this.dataTable.fnSettings();
            for(iCol = 0; iCol < oSettings.aoPreSearchCols.length; iCol++) {
                oSettings.aoPreSearchCols[ iCol ].sSearch = '';
            }
            oSettings.oPreviousSearch.sSearch = '';
            this.dataTable.fnDraw();
        },

        /**
         * Redraws the datatable
         */
        refresh: function() {
            this._cleanView();
            this.initialize();
        },

        /**
         * Event handler to preform tests refresh after test execution have been added
         * */
        _updateTestsHandler: function (e, data) {
            if (data && data.statusBar) {
                this._updateExecutionStatus(data.statusBar);
            }
            this.refresh();
        },

        /**
         * Handles the creation of the datatable.
         */
        _createTestsDataTable: function() {
            var self = this;

            var path = contextPath + '/rest/raven/1.0/testplan/' + this.testPlanKey + "/test";

            if(this.dataTable) {
                this.dataTable.fnDestroy();
            }

            this.dataTable = $('#plan-entries-table').xrayDataTable({
                "oLanguage": {
                    "oPaginate": {
                        "sFirst": AJS.I18n.getText('common.label.first'),
                        "sLast": AJS.I18n.getText('common.label.last'),
                        "sNext": AJS.I18n.getText('common.label.next'),
                        "sPrevious": AJS.I18n.getText('common.label.previous')
                    },
                    "sInfoEmpty": AJS.I18n.getText('view-issue-testplan-tests-section.infoempty'),
                    "sEmptyTable": AJS.I18n.getText('view-issue-testplan-tests-section.emptytable'),
                    "sInfo": AJS.I18n.getText('common.table.label.showing-records'),
                    "sZeroRecords": AJS.I18n.getText('common.table.label.no-records-found'),
                    "sInfoFiltered": AJS.I18n.getText('common.table.label.table-filtered-from'),
                    "sSearch":  AJS.I18n.getText('common.table.label.search'),
                    "sLengthMenu": AJS.I18n.getText('common.table.label.show-menu-entries')
                },
                "sDom" : "l<t>ip",
                "bServerSide" : true,
                "sAjaxSource": path,
                "sAjaxDataProp": "entries",
                "bProcessing": false,
                "bAutoWidth": false,
                "sPaginationType": "full_numbers",
                "bSort": true,
                "aaSorting": [],
                "aoColumnDefs": this._createDataTableColumnDefs(),
                "fnServerData": function( sSource, aoData, fnCallback, oSettings){
                    self._fnServerData( sSource, aoData, fnCallback, oSettings );
                    if (self.bulkEditView) {
                        self.bulkEditView.setCheckBoxRowsSelectionValue(false);
                    }
                },
                "fnServerParams": function ( aoData ) {
                    aoData.push({"name": "analysisEnvironment", "value": self.analysisEnvironment});
                },
                "fnRowCallback": function( nRow, aData, iDisplayIndex, iDisplayIndexFull ) {
                    nRow.id = aData["key"];
                    $(nRow).data("assignee", aData["assignee"]);
                    $(nRow).addClass("testentry");

                    return nRow;
                },
                "fnDrawCallback": function( oSettings ) {
                    // set initial overflow to allow overflow columns to be visible
                    self.$el.css("overflow", "initial");

                    // hack for drag&drop to work
                    $('#plan-entries-table').closest('div').addClass('datatable-drag-overflow');

                    self._toogleTestsTable();
                    if (self.bulkEditView) {
                        self.bulkEditView.showHideBulkActions();
                    }

                    // remove loose tipsies
                    $('.tipsy:has(.testsets-custom-field-tooltip-close)').remove();

                    //add tipsy for non-executable tests
                    $(".not-executable-test-testplan").tipsy();

                    if(self.innerTableIds) {
                        self.innerTableIds.forEach(function(innerTableId) {
                            var imageUrlBase = contextPath + '/download/resources/com.xpandit.plugins.xray:xray-project-tab-test-panel/images/datatables/';
                            $("img#"+innerTableId).attr("src" , imageUrlBase + "expand_enabled.png");
                            var innertableview = new RAVEN.TestPlan.InnerTableView({test:innerTableId, environments:self.filtersView.lastTestEnvironmentsValue});
                        });
                    }
                },
                "fnInitComplete": function(settings, json) {
                    $(".testplan-actions-icon").parent().css("min-width","114px");
                    $(".testplan-actions-icon").parent().css("white-space","nowrap");
                }
            });

            if (this.canEdit === true) {
                this._initSortableTable("#plan-entries-table", false);
            }

            this._initSortableColumns();
        },

        /**
         * Creates the DataTable column def descriptors dynamically based on the user column layout
         */
        _createDataTableColumnDefs: function() {
            var self = this;
            var columns = _.values(this.model.attributes);

            var bulkToogleColumnTitle = null;
            var bulkActionColumnTitle = null;
            if (self.canEdit === true) {
                bulkToogleColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkSelectAllInput();

                var bulkActions = RAVEN.TestPlan.testBulkMenuActions({ 'canEdit' : self.canEdit});
                bulkActionColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkActionTrigger({id: "tests-bulk-action",actions: bulkActions});
            }

            var targets = 0;
            var firstColumnDefs = [
                {
                    "mData": null,
                    "sTitle": bulkToogleColumnTitle,
                    "mDataProp": null,
                    "sWidth": "64px",
                    "sClass": "drag-drill-td",
                    "sDefaultContent": "<input type='checkbox' ></input>",
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                        if (self.canEdit === true) {
                            return RAVEN.DataTable.BulkEdit.renderBulkSelectInputAndExpand({contextPath: contextPath,id: row["key"], draggable: row["draggable"]});
                        } else {
                            return RAVEN.DataTable.BulkEdit.renderExpand({contextPath: contextPath,id: row["key"]});
                        }
                    },
                    "aTargets": [ targets ]
                }
            ];

            if(!this.isTestRepositoryEnabled) {
                targets += 1;
                firstColumnDefs.push({
                    "mData": "rank",
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                    if(row["draggable"]){
                        return "<span class='rank' style='padding-left: 2px'>" + data + "</span>";
                    }
                    return "";
                },
                    "aTargets": [ targets ]
                });
            }

            var userColumnDefs = this._createDataTableUserColumn(columns, targets + 1);
            var lastColumnDefs = [
                {
	                "mData": "latestStatus",
	                "sTitle": "Latest Status",
	                "sWidth": "110px",
                    "bSortable": false,
	                "mRender": function ( data, type, row ) {
	                    return data;
	                },
	                "aTargets": [ targets + 1 + columns.length ]
                },
                {
                    "mData": null,
                    "sTitle": bulkActionColumnTitle,
                    "sWidth": "114px",
                    "mRender": function ( data, type, row ) {
                        var hasStartedExecuting = row["startedExecuting"];
                        var canExecuteTestRun = row["canExecute"];
                        var html = '';

                    	if(self.canEdit == true) {
                    		html =  RAVEN.TestPlan.testMenu({
                                key: row["key"],
                                canEdit: self.canEdit,
                                rank: row["rank"],
                                draggable: row["draggable"]
	                        });
						}

                        if(canExecuteTestRun === false) {
                            var str = row["disallowedExecuteMessage"];
                            html += '<a class="not-executable-test-testplan" title="'+str +'" style="margin-right:48px;margin-left:0px;"></a>';
                        }
                    	return html;
                    },
                    "bSortable": false,
                    "bSearchable": false,
                    "aTargets": [ targets + 2 + columns.length]
                }
            ];

            return firstColumnDefs.concat(userColumnDefs).concat(lastColumnDefs);
        },

        /**
         * Creates the DataTable column def for user columns
         */
        _createDataTableUserColumn: function(columns, previousIndex) {
            return _.map(columns, function(column, index) {
                return {
                    "mData": 'userColumns.' + column.id,
                    "bSortable": false,
                    "bSearchable": false,
                    "aTargets": [ previousIndex + index ]
                };
            });
        },

        message: function (title, message, type) {
            type = (type === undefined || type == "") ? "error" : type;

            if (!message && title) {
                message = title;
                title = null;
            }

            if (type === "error") {
                this._clearMessage();
                AJS.messages.error('#raven-plan-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "warning") {
            	this._clearMessage();
    			AJS.messages.warning('#raven-plan-tests-message-bar', {
    				title : title,
    				body : message
    			});
    		}

            if (type === "success") {
                AJS.messages.success('#raven-plan-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "info") {
                AJS.messages.info('#raven-plan-tests-message-bar', {
                    title: title,
                    body: message
                });
            }
        },

        _clearMessage: function() {
            $('#raven-plan-tests-message-bar').empty();
        },

        _fnServerData: function( sSource, aoData, fnCallback, oSettings ) {
            var self = this;

            if (this.filters) {
                aoData = aoData.concat(this.filters);
            }

            $('#raven-plan-tests').block();
            oSettings.jqXHR = $.ajax( {
                "dataType": 'json',
                "type": "GET",
                "url": sSource,
                "data": aoData,
                "error": function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }
                    self.message(AJS.I18n.getText('view-issue-testplan-tests-section.processingerror'), errorThrown, 'error');
                },
                "complete": function() {
                    $('#raven-plan-tests').unblock();
                },
                "success": fnCallback
            });
        },

        _updateHistory: function() {
            if ($('#changehistory-tabpanel').attr('class') == 'active') {
                var alt_token = $('input[name="atl_token"]').val();
                $.ajax({
                    type : "GET",
                    url : contextPath + '/browse/' + this.testPlanKey +"?page=com.atlassian.jira.plugin.system.issuetabpanels:changehistory-tabpanel&_pjax=true&_="+alt_token,
                    contentType : "text/plain",
                    success : function (data, status) {
                        var tab = $(data).find('#activitymodule div.mod-content #issue_actions_container');
                        var jiraTab = $('#activitymodule div.mod-content').find('#issue_actions_container');
                        $(jiraTab).html(tab);
                    }
                });
            }
        },

        _updateDates: function () {
            var updatedDate = null;

            if(AJS.$('#datesmodule #updated-date').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-date time.livestamp');
            } else if(AJS.$('#datesmodule #updated-val').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-val time.livestamp');
            } else {
                return null;
            }

            $.ajax({
                type : "GET",
                url : AJS.contextPath() + '/rest/raven/1.0/date/' + JIRA.Issue.getIssueKey() + "/updated",
                contentType : "application/json",
                success : function (data) {
                    var title = data.title;
                    var dateTimeIso8601 = data.dateTimeIso8601;
                    updatedDate.attr("datetime", dateTimeIso8601);
                    $.each(AJS.$('#datesmodule dd.date'), function() {
                        if($(this).find('#updated-date time.livestamp').length > 0 || $(this).find('#updated-val time.livestamp').length > 0) {
                            $(this).attr("title", title);
                        }
                    });
                    updatedDate.livestamp();
                }
            });
        },

        _setEditable: function(editable) {
            if (editable === true)
                $('#add-tests').show();
            else
                $('#add-tests').hide();
        },

        _setExecutable: function(executable) {
            if (executable === true) {
                $('.execute-command').show();
            } else {
                $('.execute-command').hide();
            }
        },

        _setEnvironments: function() {
            var environments = $('#raven-testplan-testexec-environments').val();
            // remove first and last characters "[", "]"
            environments = environments.substring(1, environments.length-1);
            // string to array
            this.environments = environments.split(", ");
        },

        _toogleTestsTable: function() {
            if (this.dataTable && this.dataTable.fnRecordsTotal() > 0) {
                $('#raven-plan-tests .notests').hide();
                $('#plan-tests-section').show();
                $('#add-testexecs').show();
            } else {
                $('#plan-tests-section').hide();
                $('#raven-plan-tests .notests').show();
                $('#add-testexecs').hide();
            }
        },

        _addTests: function(testKeys, assignee) {
            var self = this;

            $('#raven-plan-tests').block();

            var path = contextPath + '/rest/raven/1.0/testplan/' + this.testPlanKey + '/test';

            if (!assignee || assignee == "-1") {
                assignee = null;
            }

            var value = {
                "keys": testKeys,
                "assignee": assignee
            };
            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                timeout: 60000, // sets timeout to 60 seconds
                data: JSON.stringify(value),
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-testplan-tests-section.add.error', self.testPlanKey), errorThrown);
                    $('#raven-plan-tests').unblock();
                },
                success: function(data, textStatus, jqXHR) {
                    if (data.errors && data.errors.length > 0) {
                        var errorMsg = '';
                        $.each($(data.errors), function(i, msg) {
                            errorMsg = errorMsg + msg;
                            if (i < data.errors.length) {
                                errorMsg = errorMsg + '<br/>';
                            }
                        });
                        self.message(AJS.I18n.getText('view-issue-testplan-tests-section.add.controlled.error', self.testPlanKey), errorMsg);
                    }

                    $('#raven-plan-tests').unblock();
                    self._processAddResults(data);
                },
                complete: function() {
                    self._updateHistory();
                    self._updateDates();
                }
            });
        },

        _processAddResults: function(data) {
            // new tests
            if (data.tests && data.tests > 0) {
                this.dataTable.fnDraw();
            }

            if(data.statusBar) {
            	this._updateExecutionStatus(data.statusBar);
            }
        },

        _updateExecutionStatus: function(statusbar) {
            var statusContent = RAVEN.TestPlan.planProgressBar({statusbar: statusbar});
            var pb = $('#plan-tests-progressbar');
            pb.empty();
            pb.append(statusContent);
        },

        _deleteTest: function(testKeys) {
            var self = this;

            $('#raven-plan-tests').block();

            //if we have records to delete
            if(testKeys && testKeys.length > 0) {
                //Build query param
                var testKeysParam = this._buildQueryParamsFromSelectedRows(testKeys);

                $.ajax({
                    type: "DELETE",
                    url: contextPath + '/rest/raven/1.0/testplan/' + self.testPlanKey + "/test?" + testKeysParam + "" ,
                    processData: true,
                    async: true,
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }
                        if (testKeys.length == 1) {
                            self.message(AJS.I18n.getText('view-issue-testplan-tests-section.delete.error', testKeys[0]), errorThrown);
                        } else {
                            self.message(AJS.I18n.getText('view-issue-testplan-tests-section.bulk.delete.error'), errorThrown);
                        }
                        $('#raven-plan-tests').unblock();
                    },
                    success: function(data, textStatus, jqXHR) {
                        $('#raven-plan-tests').unblock();
                        self._deleteTestsCallback(testKeys, data);
                    },
                    complete: function() {
                        self._updateHistory();
                        self._updateDates();
                    }
                });
            }
        },

        _deleteTestRun: function(testKey, testExecKey) {
            var self = this;

            $('#raven-plan-tests').block();

                $.ajax({
                    type: "DELETE",
                    url: contextPath + '/rest/raven/1.0/testplan/' + self.testPlanKey + '/testexec/' + testExecKey + "/test?testKeys=" + testKey+"",
                    processData: true,
                    async: true,
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }

                        self.message(AJS.I18n.getText('view-issue-testexec-tests-section.delete.error', testKey), errorThrown);
                        $('#raven-plan-tests').unblock();
                    },
                    success: function(data, textStatus, jqXHR) {
                        $('#raven-plan-tests').unblock();
                        self._deleteTestRunCallback(testKey, testExecKey, data.latestStatus);
                        self._updateExecutionStatus(data.planStatusBar);
                    },
                    complete: function() {
                        self._updateHistory();
                        self._updateDates();
                    }
                });
        },

        _deleteTestsCallback: function(testKeys, data) {
            var self = this;
            if (testKeys && testKeys.length > 0) {
                var allData = this.dataTable.fnGetData();
                if (allData.length >= 1) {
                    _.each(_.initial(testKeys), function(testKey) {
                        self.dataTable.fnDeleteRow(testKey, null, false);
                    });
                    this.dataTable.fnDeleteRow(_.last(testKeys));
                } else {
                    // don't refresh the table if the last element is removed because it will request more data from server
                    this.dataTable.fnDeleteRow(_.last(testKeys), null, false);
                }
            }
            if(data) {
            	this._updateExecutionStatus(data);
            }

        },

        _deleteTestRunCallback: function(testKey, testExecKey, latestStatus) {
        	// remove row
        	$("tr[data-test-key='"+testKey+"'][data-exec-key='"+testExecKey+"']").remove()
        	// check if table is empty
        	var tableSize = $("td#"+testKey+" table tbody tr").length;

        	if(tableSize == 0) {
        		$("td#"+testKey).html(AJS.I18n.getText('view-issue-testplan-tests-section.innertable.notestexecs', testKey))
        	}

        	// update number of TEs
        	if($("#column-test-plan-entry-column-testexec_count").length > 0){
        		if(tableSize == 0) {
        			$("#"+testKey + " .test-exec-count").parent().html(0);
        		} else {
        			$("#"+testKey + " .test-exec-count").html(tableSize);
        		}
        	}

        	// update status from Test
        	var statusDiv = RAVEN.TestPlan.statusDiv({status:latestStatus});
        	var classes = $(statusDiv).attr("class");
        	var testStatusDiv = $(statusDiv).attr("class",classes + " teststatus");
        	$("tr#"+testKey+" .teststatus").parent().html(testStatusDiv);
        },

        _moveTest: function(table, movedTestData, ancorTestData, position) {
            var self = this;

            $('#raven-plan-tests').block();

            var path = contextPath + '/rest/raven/1.0/testplan/' + this.testPlanKey + '/test/' + movedTestData.key + '/move'

            var value = {};
            if (position == AFTER) {
                value.after = '/' + ancorTestData.key;
            } else if (position == BEFORE) {
                value.before = '/' + ancorTestData.key;
            } else {
                value.index = position;
            }
            var data = JSON.stringify(value);

            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                data: data,
                async: true,
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }
                    self.message(AJS.I18n.getText('view-issue-testplan-tests-section.move.error', movedTestData.key), errorThrown);

                    $('#raven-plan-tests').unblock();
                },
                success: function(data, textStatus, jqXHR) {
                    $('#raven-plan-tests').unblock();
                    table.fnDraw(false);
                }
            });
        },

        _initSortableColumns: function() {
            var self = this;

            $('#plan-entries-table thead tr').sortable({
                items: 'th:not(.fixed)',
                placeholder: "dragging-placeholder",
                update: function (event, ui) {
                    var ancorNode = ui.item.prev();
                    var columnId = $(ui.item).data('column-id');
                    if (columnId) {
                        var anchorColumnId = null;
                        if (!$(ancorNode).hasClass('fixed')) {
                            anchorColumnId = $(ancorNode).data('column-id');
                        }

                        self.columnLayoutView.moveColumn(columnId, anchorColumnId);
                    }
                }
            });
        },

        /**
         * Adds support for moving datatable items
         */
        _initSortableTable: function(tableSelector) {
            var self = this;

            $(tableSelector).sortable({
                items:'tbody tr.testentry',
                handle:'.draghandle',
                placeholder: "drag-placeholder",
                helper: function(e, tr) {
                    self._closeViewExtender(tr[0]);
                    var $originals = tr.children();
                    var $helper = tr.clone();
                    $helper.children().each(function(index) {
                        // Set helper cell sizes to match the original sizes
                        $(this).width($originals.eq(index).width())
                    });
                    return $helper;
                },

                update: function (event, ui) {
                    var position = null;
                    var ancorNode = ui.item.prev();

                    if (ancorNode.length > 0) {
                        position = AFTER;
                    } else {
                        ancorNode = ui.item.next();
                        if (ancorNode.length > 0) {
                            position = BEFORE;
                        }
                    }

                    if (position !== null) {
                        var movedTestData = self.dataTable.fnGetData(ui.item[0]);
                        var ancorTestData = self.dataTable.fnGetData(ancorNode[0]);

                        if(!ancorTestData && $(ancorNode[0])[0] && $(ancorNode[0])[0].dataset) {
                            ancorTestData = {key : $(ancorNode[0])[0].dataset.testKey};

                            if(!ancorTestData.key) {
                                ancorTestData = {key : $(ancorNode[0]).prev().attr('id')};
                            }
                        }

                        if (movedTestData && movedTestData.key && ancorTestData && ancorTestData.key) {
                            self._moveTest(self.dataTable, movedTestData, ancorTestData, position);
                        }
                    }
                }
            });
        },

        /**
         * Inline execution of Test Runs.
         */
        _inlineExecuteTestRun: function($row, statusId) {
            var self = this;

            $('#raven-plan-tests').block();
            var currentStatus = $($row[0]).attr("data-current-status");
            if (currentStatus !== statusId) {
                var testKey = $($row[0]).data("test-key");
                var testExecKey = $($row[0]).data("exec-key");

                var path = contextPath + '/rest/raven/1.0/testplan/' + this.testPlanKey + '/execute/' + testExecKey + '/' + testKey;

                $.ajax({
                    type: "POST",
                    url: path,
                    processData: true,
                    async: true,
                    data: statusId.toString(),
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }

                        self.message(AJS.I18n.getText('view-issue-testexec-tests-section.inline-execute.error', self.testExecKey), errorThrown);
                    },
                    success: function(data, textStatus, jqXHR) {
                    	$('#raven-plan-tests').block();
                    	$("tr[data-test-key='"+testKey+"'][data-exec-key='"+testExecKey+"']").attr("data-current-status",data.obj.status.id);

                    	// update TE Status
                    	var statusDiv = RAVEN.TestPlan.statusDiv({status:data.obj.status});
                    	$("tr[data-test-key='"+testKey+"'][data-exec-key='"+testExecKey+"'] .planStatus").html(statusDiv);

                    	// update TE TestRunMenu
                    	var testRunMenu = RAVEN.TestPlan.testRunMenu({
                    		key: data.obj.issueKey,
                    		testkey: testKey,
                    		canEdit: self.canEdit,
                    		canAssign: self.canAssign,
                    		canExecuteTest: data.obj.canExecuteTest,
                    		hasStartedExecuting: data.obj.hasStartedExecution,
                    		canExecuteInline: self.canExecuteInline,
                    		executeHref: data.obj.executeHref,
                    		transitions: data.obj.transitions,
                    		isFinal: data.obj.status.isFinal,
            				isStatusExecutable: data.obj.canExecuteTestExecution,
            				disallowedMessage: data.obj.disallowedExecuteMessage
                    	});

                        var actionsMenu = RAVEN.TestPlan.actionsMenu({
                            key: data.obj.issueKey,
                            testKey: testKey,
                            canEdit: self.canEdit,
                            canAssign: self.canAssign
                        });

                    	$("tr[data-test-key='"+testKey+"'][data-exec-key='"+testExecKey+"'] .planTestRunMenu").html(actionsMenu + testRunMenu);

                    	// update Test Status
                    	var statusDiv = RAVEN.TestPlan.statusDiv({status:data.obj.newStatus});
                    	var classes = $(statusDiv).attr("class");
	                    var testStatusDiv = $(statusDiv).attr("class",classes + " teststatus");
	                    $("tr#"+testKey+" .teststatus").parent().html(testStatusDiv);

                    	// update status bar
                    	self._updateExecutionStatus(data.obj.statusBar);
                    },
                    complete: function() {
                        $('#raven-plan-tests').unblock();
                    }
                });
            }
        },

        /**
         * Build query param
         */
        _buildQueryParamsFromSelectedRows: function(selectedIds) {
            var testKeys = "";
            var i;
            for (i = 0; i < selectedIds.length; ++i) {
                if(testKeys.length > 0){
                        testKeys+="&";
                }
                testKeys += "testKeys=" + selectedIds[i];
            }

            return testKeys;
        },

        _getTestExecHref: function(testExecKey) {
            var link = contextPath + '/browse/' + testExecKey;
            return link;
        },

        _confirmationDeleteTest: function(el, yesFunction, bulk) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var title = null;
                    var message = null;
                    if (bulk === true) {
                        title = AJS.I18n.getText('view-issue-testplan-tests-section.bulk.delete.confirm.title');
                        message = AJS.I18n.getText('view-issue-testplan-tests-section.bulk.delete.confirm.message', self.testPlanKey);
                    } else {
                        title = AJS.I18n.getText('view-issue-testplan-tests-section.delete.confirm.title', el);
                        message = AJS.I18n.getText('view-issue-testplan-tests-section.delete.confirm.message', self.testPlanKey);
                    }
                    var severity = "info";
                    var innerHtml = RAVEN.Dialog.deleteConfirmationDialog({title:title, message:message, severity:severity});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(el);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        _confirmationDeleteTestRun: function(testKey, testExecKey, yesFunction) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var title = null;
                    var message = null;
                    title = AJS.I18n.getText('view-issue-testplan-tests-section.delete.confirm.title', testKey);
                    message = AJS.I18n.getText('view-issue-testexec-tests-section.delete.confirm.message', testExecKey);
                    var severity = "info";
                    var innerHtml = RAVEN.Dialog.deleteConfirmationDialog({title:title, message:message, severity:severity});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(testKey, testExecKey);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        _renderBulkModeButton: function() {
            var $wrapper = this.$el.find('.xrayDataTables_wrapper');
            // if filter already exists then return
            if ($wrapper.find('#bulk-edit-action').length > 0) {
                return;
            }

            this.bulkEditView = new RAVEN.DataTable.BulkEditView({
                el: $wrapper
            });

        },

        _updateRankRow: function(newRankValue, originalRankSpanValue, row) {
            //Validate values
            if(newRankValue > 0 && newRankValue <= this.dataTable.fnSettings().fnRecordsTotal() && newRankValue != originalRankSpanValue) {
                var movedTestData = this.dataTable.fnGetData(row);

                if (movedTestData && movedTestData.key) {
                    this._moveTest(this.dataTable, movedTestData, null, newRankValue);
                }
            }
        },

        _renderTestsToolbarView: function() {
            var $wrapper = this.$el.find('.xrayDataTables_wrapper');
            // if filter already exists then return
            if ($wrapper.find('#toolbar-wrapper').length > 0) {
                return;
            }

            var toolbarHtml = RAVEN.TestPlan.renderTestsToolBar({environments: this.environments});
            $wrapper.find('.xrayDataTables_length').after(toolbarHtml);

            this.columnLayoutView = new RAVEN.TestPlan.Tests.TestsColumnLayoutView({
                mainView: this
            });
        },

        _assignTestRun: function(username, testKey, testExecKey) {
            var self = this;

            $('#raven-plan-tests').block();

            var path = contextPath + '/rest/raven/1.0/testexec/' + testExecKey + '/assign';

            var value = {
                keys: [testKey],
                assignee: !username || username == -1 ? null : username
            };
            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                data: JSON.stringify(value),
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-testexec-tests-section.assign.error'), errorThrown);
                },
                success: function(data, textStatus, jqXHR) {
                    self._onAssignOperationSuccess(data, testKey, testExecKey);
                },
                complete: function() {
                    $('#raven-plan-tests').unblock();
                    self._updateHistory();
                    self._updateDates();
                }
            });
        },

        _onAssignOperationSuccess: function(data, testKey, testExecKey) {
        	$("tr[data-test-key='"+testKey+"'][data-exec-key='"+testExecKey+"']").data("test-run-assignee",data[0].assignee);
        },

        _createSubtleSingleSelect: function(selectField) {
            var items = _.map(selectField.find('option'), function(item) {
                return {
                    'id' : item.value,
                    'name' : item.text,
                };
            });

            var selectedItem = _.first(_.map(selectField.find('option:selected'), function(item) {
                return {
                    'id' : item.value,
                    'name' : item.text,
                };
            }));

            var subtlePicker = new RAVEN.Field.SubtleSingleSelect(items, selectedItem, selectField.closest('#raven-subtle-container'));
            subtlePicker.render();

            // remove select element. This is only for retrieving the initial items.
            selectField.remove();

            return subtlePicker;
        },

        /****************************************************************
         *                        EVENT HANDLERS
         ****************************************************************/

        _delegateUnscopedEvents: function() {
            var self = this;
            $(document).off('.raven.testplan');

            $.each(this.unscopedEvents, function(key, val) {
                var eventName = key.substring(0, key.indexOf(' '));
                var eventTrigger = key.substring(key.indexOf(' ') + 1);

                $(document).on(eventName, eventTrigger, $.proxy(self[val], self));
            });
        },

        _undelegateUnscopedEvents: function() {
            $(document).off('.raven.testplan');
        },

        /**
         * Add tests button click handler
         */
        _onAddTestsClick: function() {
            var self = this;

            this._clearMessage();

            var labels = {};
            labels.label = AJS.I18n.getText('dialogs.issueselector.test.label');
            labels.title = AJS.I18n.getText('dialogs.issueselector.test.title', this.testPlanKey);
            labels.description = AJS.I18n.getText('dialogs.issueselector.test.description');
            labels.enter = AJS.I18n.getText('dialogs.issueselector.test.enter');

            var issueTypeId = $('#test-issue-type-id').val();

            new RAVEN.Dialogs.createIssueSelectorAdvancedDialog('TEST', 'key not in testPlanTests(' + this.testPlanKey + ')', issueTypeId, this.testPlanKey, labels, function(issues, assignee) {
                if (issues && issues.length > 0) {
                    self._addTests(issues, assignee);
                }
            }, null, self.canAssign);
        },

        /**
         * Add test sets button click handler
         */
        _onAddTestSetsClick: function() {
            var self = this;

            this._clearMessage();

            var labels = {};
            labels.label = AJS.I18n.getText('dialogs.issueselector.testset.label');
            labels.title = AJS.I18n.getText('dialogs.issueselector.testset.title', this.testPlanKey);
            labels.description = AJS.I18n.getText('dialogs.issueselector.testset.description');
            labels.enter = AJS.I18n.getText('dialogs.issueselector.testset.enter');

            var issueTypeId = $('#testset-issue-type-id').val();

            new RAVEN.Dialogs.createIssueSelectorAdvancedDialog('TESTSET', 'key not in testSetFullyIn(' + this.testPlanKey + ')', issueTypeId, this.testPlanKey, labels, function(issues, assignee) {
                if (issues && issues.length > 0) {
                    self._addTests(issues, assignee);
                }
            }, null, self.canAssign);
        },

        /**
         * Delete test click handler
         */
        _onTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            var testKey = $(e.target).closest('a').data('key');
            this._confirmationDeleteTest([ testKey ], this._deleteTest);
        },

        /**
         * Delete test run click handler
         */
        _onTestRunDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            var testExecKey = $(e.target).closest('a').data('exec-key');
            var testKey = $(e.target).closest('a').data('key');
            this._confirmationDeleteTestRun(testKey, testExecKey, this._deleteTestRun);
        },

        /**
         * Delete Bulk test click handler
         */
        _onBulkTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            if(this.bulkEditView.getSelectedRowIds().length > 0) {
                this._confirmationDeleteTest(this.bulkEditView.getSelectedRowIds(), this._deleteTest, true);
            }
        },

        /**
         * Creates Test Execution with selected tests
         */
        _onBulkCreateTestExecutionClick: function(e) {
            e.stopPropagation();
            e.preventDefault();
            if(this.bulkEditView.getSelectedRowIds().length > 0) {
                this._createTestExecutionDialog(this.bulkEditView.getSelectedRowIds());
            }
        },

        _onCreateTestExecutionClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._createTestExecutionDialog([]);
        },

        _onCreateTestExecutionWithStatusClick: function(e) {
        	e.stopPropagation();
            e.preventDefault();
            var self = this;
            var statuses;

			$.ajax({
				type: "GET",
				url: contextPath +'/rest/raven/1.0/testplan/' + this.testPlanKey + '/getAllStatus',
				async: false,
				contentType: "application/json",
				success: function(data) {
					statuses = data
				}
			});

			 var labels = {};
	         labels.title = AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.select-status.title');
	         labels.enter = AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.create-new.enter');

	         new RAVEN.Dialogs.pickStatusesDialog(labels, this.testPlanKey, statuses, function(pickedStatus) {
	        	 self._createTestExecutionDialog([],pickedStatus);
	         });

        },

        _messageFlag: function(type, text, close) {
            return require('aui/flag')({
                type: type,
                body: text,
                close: close
            });
        },

        _createTestRuns: function(issueKey) {
            if (!this.flagCreatedTestExec) {
                return;
            }
            this.flagCreatedTestExec = false;
            var path = contextPath +'/rest/raven/1.0/testplan/' + this.testPlanKey + '/createtestruns?testexec=' + issueKey;
            var self = this;
            $.ajax({
                type: "PUT",
                url: path,
                async: true,
                contentType: "application/json",
                timeout: 960000,

                error: function (qXHR, textStatus, errorThrown) {
                    self._messageFlag('error', errorThrown, 'false');
                },
                success: function () {
                    self._messageFlag('success', AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.create-new.finished',
                        '<a href=' + contextPath + '/browse/' + issueKey + ' class="link-title">' + issueKey + '</a>',
                        'false'));
                },
                complete: function() {
                    if (self.warningMessage) {
                        self.warningMessage.close();
                    }
                }
            });
        },

        _createTestExecutionDialog: function(tests,pickedStatus) {
            var self = this;
            var projectKey;
            var projectId;
            var assignee;
            var fixVersions;
            var customField;
            var customFieldValues;
            self.dialogLoaded = false;

            var path;
            if(tests.length > 0){
            	path = contextPath +'/rest/raven/1.0/testplan/' + this.testPlanKey + '/createtestexechelper?testKeys=' + tests;
            } else if( pickedStatus && pickedStatus.length > 0 ) {
            	path = contextPath +'/rest/raven/1.0/testplan/' + this.testPlanKey + '/createtestexechelper?selectedStatus=' + pickedStatus;
            } else {
            	path = contextPath +'/rest/raven/1.0/testplan/' + this.testPlanKey + '/createtestexechelper';
            }

            $.ajax({
				type: "GET",
				url: path,
				async: false,
				contentType: "application/json",
				success: function(data) {
					assignee = data.assignee;
					projectKey = data.project;
					projectId = data.projectId;
					fixVersions = data.fixVersion;
					customField = data.customField;
					customFieldValues = data.customFieldValues;

					if(tests && tests.length === 0 ) {
						tests = data.tests;	// can i not send this in the path and just keep it here in the JS
					}
				}
			});

			self.tests = tests;
            var labels = {};
            labels.title = AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.create-new.title', this.testPlanKey);
            labels.enter = AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.create-new.enter');
            var summary = AJS.I18n.getText('view-issue-testplan-section.create-testexec.summary', this.testPlanKey);

            if($('#has-required-fields').val() !== "true"){
                new RAVEN.Dialogs.testPlanTestExecDialog(labels, projectKey, assignee, summary, fixVersions, customField, function(projectKey, summary, assignee, canRedirectToTestExecPage, fixVersion, environment, revision) {
                    self._createTestExec(tests, projectKey, summary, assignee, canRedirectToTestExecPage, fixVersion, environment, revision);
                });
            }else {
                var testExecIssueType = parseInt($('#testexec-issue-type-id').val());
                var createForm = JIRA.Forms.createCreateIssueForm({
                    issueType: testExecIssueType
                });

                // select trigger for create issue in global context
                var dialog = createForm.asDialog({
                    id: 'raven-create-exec-dialog'
                });

                createForm.bind("sessionComplete", function (e, data) {
                    if (data && data.length > 0) {
                        var issue = data[data.length-1]; //get the last issue created
                        if (issue) {
                        	var title = AJS.I18n.getText('view-issue-testplan-section.action.add.controlled.success', issue.issueKey);
                        	var message = AJS.I18n.getText('view-issue-testplan-section.action.add.success', contextPath, issue.issueKey, issue.summary, self.testPlanKey, $("#summary-val").html().split("<")[0]);

                        	self.message(title, message, "success");
                            self.refresh();
                        }
                    }
                });

                self.fixVersions = fixVersions;

                if (tests.length > 100) {
                    this.flagCreatedTestExec = true;
                    this.warningMessage = this._messageFlag('info', AJS.I18n.getText('view-issue-testplan-section.action.associate-test-execution.create-new.warning'));
                }
                dialog.show();

                JIRA.bind(JIRA.Events.NEW_CONTENT_ADDED, function(e, context, reason) {

                    var dialog = $('#raven-create-exec-dialog.jira-dialog-open, #raven-create-exec-dialog.aui-dialog-open');

                    context.find('.qf-create-another').hide();

                    if (dialog.length > 0 && !self.dialogLoaded) {
                    	self.dialogLoaded = true;
                    	var testExecCF = $('#raven-testexec-test-customFieldId').val();
                    	var testPlanCF = $('#raven-testplan-customFieldId').val();
                        // add Test

                        if (self.tests) {
                            var $selector = context.find('select[id=' + testExecCF + ']');
                                self.tests.forEach(function (element) {
                                    $selector.trigger('selectOption', [{value: element}]);
                                }, this);
                        }

                        context.find('select[id='+testPlanCF+']').trigger('selectOption',[{label:self.testPlanKey,value:self.testPlanKey,title:self.testPlanKey}]);

                        // Fix Versions
                        if(self.fixVersions !== null && self.fixVersions !== undefined) {
                            var $fixVersions = $(dialog).find("#fixVersions");
                            $(self.fixVersions).each(function (index, version) {
                                context.find('select[id="fixVersions"] option').each(function () {
                                    $fixVersions.trigger('selectOption', [{
                                        value: version.value,
                                        title: version.name,
                                        label: version.name
                                    }]);
                                });
                            });
                        }

                        // add assignee

                        if(assignee !== 'undefined' && assignee !== "") {
	                        var assigneeObj;
	                        $.ajax({
	                            type: "GET",
	                            url: contextPath +'/rest/api/2/user?key=' + assignee,
	                            async: false,
	                            contentType: "application/json",
	                            success: function(data) {
	                            	assigneeObj = data;
	                            }
	                        });
	                        context.find("#assignee-field").val(assigneeObj.displayName);
	                        context.find("#assignee-single-select .aui-ss-entity-icon").attr("src",assigneeObj.avatarUrls['16x16']);

	                        context.find("#assignee").append('<option value="'+assigneeObj.key+'" title="undefined" selected="selected" style="background-image: url("'+assigneeObj.avatarUrls['16x16']+'");">'+assigneeObj.displayName+'</option>')
	                        context.find("#assignee").val(assigneeObj.key);
	                        context.find('#assignee').trigger('change');
                        }

                        // add summary
                        context.find("input[id=summary]").val(AJS.I18n.getText('view-issue-testplan-section.create-testexec.summary', self.testPlanKey));

                        // add begin/end date custom fields
                        customFieldValues.forEach(function(field) {
                            var component = context.find('#' + field.cfId);
                            component.val(field.value);
                            component.trigger('change');
                        });

                        // add project needs to be last because it refreshes(locks) dialog.
                        context.find('input[id=project]').val(projectId);
                        context.find('input[id=project]').trigger('change');


                    }
                });
            }
        },

        _initTestRunEvent : function() {
            var self = this;

            JIRA.bind("QuickCreateIssue.sessionComplete", function(event, issueData) {
                if($('#has-required-fields').val() !== "true") {
                    return;
                }

                issueData.forEach(function(issueData) {
                    if (issueData.issueKey === undefined || issueData.issueKey === null) {
                        return;
                    }
                    self._createTestRuns(issueData.issueKey);
                });
            });
        },

        _populateDialog: function(e, context, reason) {

            var dialog = $('#raven-create-exec-dialog.jira-dialog-open, #raven-create-exec-dialog.aui-dialog-open');

            context.find('.qf-create-another').hide();

            if (dialog.length > 0 && !self.dialogLoaded) {
            	self.dialogLoaded = true;
            	var testExecCF = $('#raven-testexec-test-customFieldId').val();
            	var testPlanCF = $('#raven-testplan-customFieldId').val();
                // add Test

                if (tests) {
                	tests.forEach(function(element) {
                		context.find('select[id='+testExecCF+']').trigger('selectOption', [{value: element}]);
                    }, this);
                }

                context.find('select[id='+testPlanCF+']').trigger('selectOption',[{label:self.testPlanKey,value:self.testPlanKey,title:self.testPlanKey}]);

                // add assignee
                var assigneeObj;
                $.ajax({
                    type: "GET",
                    url: contextPath +'/rest/api/2/user?key=' + assignee,
                    async: false,
                    contentType: "application/json",
                    success: function(data) {
                    	assigneeObj = data;
                    }
                });

                context.find("#assignee-field").val(assigneeObj.displayName);
                context.find("#assignee-single-select .aui-ss-entity-icon").attr("src",assigneeObj.avatarUrls['16x16']);

                context.find("#assignee").append('<option value="'+assigneeObj.key+'" title="undefined" selected="selected" style="background-image: url("'+assigneeObj.avatarUrls['16x16']+'");">'+assigneeObj.displayName+'</option>')
                context.find("#assignee").val(assigneeObj.key);
                context.find('#assignee').trigger('change');

                // add summary
                context.find("input[id=summary]").val(AJS.I18n.getText('view-issue-testplan-section.create-testexec.summary', self.testPlanKey));

                // add project needs to be last because it refreshes(locks) dialog.
                context.find('input[id=project]').val(projectId);
                context.find('input[id=project]').trigger('change');
            }
        },

        _createTestExec: function(issues, projectKey, summary, assignee, canRedirectToExecutePage, fixVersion, environment, revision){
            var self = this;

            var testPlanKey = self.testPlanKey;

            $('#raven-plan-tests').block();

            var path = contextPath + '/rest/raven/1.0/testplan/' + testPlanKey + '/createtestexec?projectKey=' + projectKey + "&summary=" + summary + "&assignee=" + assignee + "&fixVersion=" + fixVersion + "&environment=" + environment + "&revision=" + revision;

            var value = JSON.stringify(issues);
            $.ajax({
                type: "POST",
                url: path,
                processData: true,
                async: true,
                timeout: 60000, // sets timeout to 60 seconds
                data: value,
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    self._clearMessage();
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }
                    self.message(AJS.I18n.getText('view-issue-testrun-section.action.add.error', testPlanKey), errorThrown);
                },
                success: function(data, textStatus, jqXHR) {
                    self._clearMessage();
                    if(data.testExecKey){
                        if(canRedirectToExecutePage){
                            window.location.href = self._getTestExecHref(data.testExecKey);
                            return;
                        }
                    }
                    if (data.errors && data.errors.length > 0) {
                        var errorMsg = '';
                        $.each($(data.errors), function(i, msg) {
                            errorMsg = errorMsg + msg;
                            if (i < data.errors.length) {
                                errorMsg = errorMsg + '<br/>';
                            }
                        });
                        self.message(AJS.I18n.getText('view-issue-testrun-section.action.add.controlled.error', testKey), errorMsg, "error");
                    }

                    // refresh view
                    if(data.testExecKey){
                    	var title = AJS.I18n.getText('view-issue-testplan-section.action.add.controlled.success', data.testExecKey);
                    	var message = AJS.I18n.getText('view-issue-testplan-section.action.add.success', contextPath, data.testExecKey, summary, testPlanKey, data.testPlanSummary);

                    	self.message(title, message, "success");
                        self.refresh();
                    }
                },
                complete: function() {
                	$('#raven-plan-tests').block();
                }
            });
        },

        /**
         * Assign Test Run click handler
         */
        _onAssignTestRunClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            var self = this;

            var $row = $(this.currentDropdownTrigger).closest('tr');

            var testKey = $row.data("test-key");
            var testExecKey = $row.data("exec-key");
            var $testRow = $("tr#"+testKey);
            var assignee = $("tr[data-test-key='" +testKey+ "'][data-exec-key='"+testExecKey+"']").data("test-run-assignee");
            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var innerHtml = RAVEN.TestExecution.renderUserPickerDialog({
                        currentAssignee: assignee,
                        issueKey: testExecKey
                    });
                    callback(innerHtml);

                    var control = new JIRA.AssigneePicker({
                        element: $('.xray-assignee-picker')
                    });
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var assignee = $('#raven-user-picker')[0].value;
                    var proxyFunc = $.proxy(self._assignTestRun, self);
                    proxyFunc(assignee, testKey, testExecKey);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        /**
         * Execute inline test click handler
         */
        _onExecuteInlineTestClick: function(e) {
            e.preventDefault();
            e.stopPropagation();

            if (this.currentDropdownTrigger) {
                var $row = $(this.currentDropdownTrigger).closest('tr');
                var statusId = $(e.target).data('status-id');

                if ($row.length > 0) {
                    this._inlineExecuteTestRun($row, statusId);
                }
            }

            return false;
        },

        /**
         * Change Test rank click handler.
         */
        _onChangeTestRankClick: function(e) {
            e.stopPropagation();
            e.preventDefault();
            this._clearMessage();

            var originalRankSpanValue = $(e.target).attr('data-key');

            var $row = $(this.currentDropdownTrigger).closest('tr');

            var rowKey = this.dataTable.fnGetData($row[0]).key;
            var label = AJS.I18n.getText('dialogs.edit-value.rank.label');
            var title = rowKey + ' - ' + AJS.I18n.getText('dialogs.edit-value.rank.title');
            this._confirmationChangeRank(title, label, $row[0], originalRankSpanValue, this._updateRankRow);
        },

        _confirmationChangeRank: function(title, message, row, originalRankSpanValue, yesFunction) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                content: function (callback) {
                    var innerHtml = RAVEN.Dialog.editValueDialog({title:title, fieldName:message, actionText:AJS.I18n.getText('common.action.ok')});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var newRankValue = $('#edit-value-dialog-value')[0].value;
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(newRankValue, originalRankSpanValue, row);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        /**
         * Save current dropdown trigger
         */
        _onDropDownTriggerClick: function(e) {
            this.currentDropdownTrigger = $(e.target);
            this.currentDropdownTrigger.closest('tr').addClass('active');
        },

        /**
         * Clear current dropdown trigger
         */
        _onDownDownHide: function() {
            if (this.currentDropdownTrigger) {
                this.currentDropdownTrigger.closest('tr').removeClass('active');
            }
        },

        _onExpandClick: function(e) {
            var self = this;
        	var imageUrlBase = contextPath + '/download/resources/com.xpandit.plugins.xray:xray-project-tab-test-panel/images/datatables/';
        	var id = e.target.id;
        	var image = $("img#"+id).attr("src");

        	if(image.indexOf("enabled") == -1 ) {
                $("img#"+id).attr("src" , imageUrlBase + "expand_enabled.png");
                var innertableview = new RAVEN.TestPlan.InnerTableView({test:id, environments:self.filtersView.lastTestEnvironmentsValue});
        	} else {
        		$("img#"+id).attr("src" , imageUrlBase + "expand_disabled.png");
        		$("#inner-"+id).remove();
        	}
        },

        _closeViewExtender: function(e) {
            var imageUrlBase = contextPath + '/download/resources/com.xpandit.plugins.xray:xray-project-tab-test-panel/images/datatables/';
            var id = e.id;
            $("img#"+id).attr("src" , imageUrlBase + "expand_disabled.png");
            $("#inner-"+id).remove();
        },

    });

    RAVEN.TestPlan.TestsView.REFRESH_VIEW = "refreshTestsView";

})(AJS.$);
