/**
 * jQuery plugin for Pretty looking right click context menu.
 *
 * Requires popup.js and popup.css to be included in your page. And jQuery, obviously.
 *
 * Usage for Xray:
 *
 *   AJS.$('.something').contextPopup({
 *     title: 'Some title',
 *     items: [
 *       {label:'My Item', action:function() { alert('hi'); }},
 *       {label:'Item #2', action:function() { alert('yo'); }},
 *       null, // divider
 *       {label:'Blahhhh', action:function() { alert('bye'); }, isEnabled: function() { return false; }},
 *     ]
 *   });
 *
 * - Joe Walnes, 2011 http://joewalnes.com/
 *   https://github.com/joewalnes/jquery-simple-context-menu
 *
 * MIT License: https://github.com/joewalnes/jquery-simple-context-menu/blob/master/LICENSE.txt
 */
jQuery.fn.contextPopup = function (menuData) {
    // Define default settings
    var settings = {
        contextMenuClass: 'contextMenuPlugin',
        linkClickerClass: 'contextMenuLink',
        headerClass: 'header',
        seperatorClass: 'divider',
        title: '',
        items: [],
        disabledMessage: ''
    };

    var NUMBER_ITEMS_ALWAYS_ENABLED = 2;
    // merge them
    AJS.$.extend(settings, menuData);

    // Build popup menu HTML
    function createMenu(e) {
        var menu = AJS.$('<ul id="context-menu-ul" class="' + settings.contextMenuClass + '"></div></ul>').appendTo(document.body);

        if (settings.title) {
            AJS.$('<li class="' + settings.headerClass + '"></li>').text(settings.title).appendTo(menu);
        }

        var countDisabledItems = 0;

        settings.items.forEach(function (item) {
            if (item) {
                var rowCode = '<li><a href="#" class="' + settings.linkClickerClass + '"><span class="itemTitle"></span></a></li>';
                var row = AJS.$(rowCode).appendTo(menu);

                var title = row.find('.itemTitle');
                title.text(item.label);

                if (item.isEnabled && !item.isEnabled()) {
                    row.addClass('disabled');
                    countDisabledItems++;
                } else if (item.action) {
                    title.attr('title', item.label);
                    row.find('.' + settings.linkClickerClass).click(function () {
                        item.action(e);
                    });
                }

            } else {
                AJS.$('<li class="' + settings.seperatorClass + '"></li>').appendTo(menu);
                countDisabledItems++;
            }

            if (countDisabledItems === (settings.items.length - NUMBER_ITEMS_ALWAYS_ENABLED)) {
                AJS.$('#context-menu-ul').attr('title', settings.disabledMessage);
            }
        });

        menu.find('.' + settings.headerClass).text(settings.title);

        return menu;
    }

    // On contextmenu event (right click)
    this.on('contextmenu', function (e) {
        var menuFrame = createMenu(e).show();

        var left = e.pageX + 5,
            /* nudge to the right, so the pointer is covering the title */
            top = e.pageY;

        if (top + menuFrame.outerHeight(true) >= AJS.$(window).height()) {
            top -= menuFrame.outerHeight(true);
        }

        if (left + menuFrame.width() >= AJS.$(window).width()) {
            left -= menuFrame.width();
        }

        // Create and show menu
        menuFrame.css({
            zIndex: 1000001,
            left: left,
            top: top
        }).on('contextmenu', function () {
            return false;
        });

        // Cover rest of page with invisible div that when clicked will cancel the popup.
        var bg = AJS.$('<div></div>')
            .css({
                left: 0,
                top: 0,
                width: '100%',
                height: '100%',
                position: 'absolute',
                zIndex: 1000000
            })
            .appendTo(document.body)
            .on('contextmenu click', function () {
                // If click or right click anywhere else on page: remove clean up.
                bg.remove();
                menuFrame.remove();
                return false;
            });

        // When clicking on a link in menu: clean up (in addition to handlers on link already)
        menuFrame.find('a').click(function () {
            bg.remove();
            menuFrame.remove();
        });

        // Cancel event, so real browser popup doesn't appear.
        return false;
    });

    return this;
};